/*
 * Author:  Konrad Rzeszutek <conradus@iname.com>
 * Date:    Tue, Sept 21, 1999
 * File:    Editor3D/DrawLine.java
 *
 * Notes:
 *
 *  This splendid class draws a nice line using the Midpoint Formula.
 *  It can draw the line as either patterned, thick, neither, or both.
 *
 * [Oct 15, 1999:]
 *   This class has been infussioned into Editor3D package. This package
 *  doesn't require line being drawn thick, patterned, and so forth.
 *  Hence, any references used for calculations of these functions has
 *  been extricted from this object (to speed it up). Also the 
 *  constructor has been modified.
 * [Oct 19, 1999:]
 *   The drawPoints has been modified not to draw negative points (can't
 *  draw them in Java coordinate system anyhow).
 *
 */

package Editor3D;

import java.awt.*;
import java.lang.*;

public class DrawLine{

    // Just a couple of constants. We use bit masking for storing info about swapping, negating, etc..
    
    final static int NOTHING  = 0;
    final static int NEGATE_X = 4;
    final static int NEGATE_Y = 2;
    final static int SWAP     = 1;

    /*
     * Initializes the DrawLine object
     *
     * require:
     *   OffG != null  && canvas != null && OffD != null
     */
    public DrawLine(Graphics OffG, Dimension OffD, int x1, int y1, int x2, int y2) {

	this.OffG = OffG;
	
	maxPoints = nPoints = 2;	// Maximum points we will ever use.

	width = OffD.width;
	height = OffD.height;
	xOrigin = width / 2;  // Our Cartesian X origin
	yOrigin = height / 2;
	xPoints = new int[maxPoints]; // Our X-points "holders"
	yPoints = new int[maxPoints]; // Our Y-points "holders"
	xPoints[0] = x1;
	xPoints[1] = x2;
	yPoints[0] = y1;
	yPoints[1] = y2; 
	this.draw();
    }
    
    /*
     * Draws the line on OffG.
     * 
     * Note:
     *  if (canAdd() || isDeleted()) then draw() will not draw anything.
     */     
    public void draw () {

	int dx, dy, incE, incNE, gM;// Used to recompute our direction.
	int[] myXPoints, myYPoints,XY; // Self-explanatory.
	
	if ((nPoints == maxPoints)) { // Just in case some foul tries to run this without all the points.
	    // Copy the points.
	    myXPoints = copy(xPoints);
	    myYPoints = copy(yPoints);

	    // Fixes the points. Look in description of fixPoints to find out exactly what it does
	    fixPoints(myXPoints, myYPoints);

	    XY = new int[2];
	    XY[0] = myXPoints[0];
	    XY[1] = myYPoints[0];
	    // Find the delta's of the new points.
	    dx = myXPoints[1] - myXPoints[0];
	    dy = myYPoints[1] - myYPoints[0];
	    incE = 2*dy;
	    incNE = 2*(dy-dx);
	     // Calculate how much we would have to increasy gM.
	    gM = 2*dy - dx;

	    while (XY[0] < myXPoints[1]) {
		if (gM <= 0) {
		    gM += incE;
		} else {
		    gM += incNE;XY[1]++;
		}
		XY[0]++; // Increase x by one.
		// makes a copy of the points, undo the points (so they
		// will confirm with Java coordinate system and then draw it.
		drawPoint(undoPoint(copy(XY))); 		
	}
	}
    }
   

    
    // A bunch of private methods used throughout the program.

    /*
      * Fixes the points for the line drawing algorhytm.
      * 
      * require:
      *   x.length() ==2 && y.length() == 2.
      *
      * return:
      *   x != xnew && y != ynew. (It changes the arguments).
      * 
      * note:
      *  this method does the followin:
      *   - changes the points to a Cartesian coordinate system.
      *   - swaps, negates x, negates y, depending in which quadrant the point
      *     resides.
      *   - that's it.
      *   - to undo the changes use undoPoint(int[]).
      */
    private void fixPoints (int[] x, int[] y) {       

	int i;     // Simple counter.
	int dx, dy;// Change over X and Y.

	if (maxPoints == nPoints) // We know we got all the points.
	    {
		// Adds/subtract values to change the x and y to a 
		// Cartesian system. xOrigin/yOrigin where calculated in the constructor.

		for (i=0; i<nPoints;i++) {
		    x[i]= x[i] - xOrigin;
		    y[i]= height - y[i] - yOrigin;
		}
		dx = x[1] - x[0];
		dy = y[1] - y[0];
		// Find in which quadrant the point and change accordincly (also set the flag)
		
		if (dx < 0) { neg(x); flag |= NEGATE_X;}
		if (dy < 0) { neg(y); flag |= NEGATE_Y; }
		if (Math.abs(dx) < Math.abs(dy)) { swap (x, y); flag|=SWAP;}

		// No more changes.
	    }
    }

    /*
      * Undo the changes that fixPoint did to X and Y's.
      * 
      * require:
      *
      *  XY.length() == 2
      *
      * return
      *  XY.length() == 2 && XYnew != XYold. (The XY will be diffrent)
      *
      * note:
      *  this method does the following:
      *   - swaps, negates x, negates y, depending in which quadrant the point
      *     reside
      *   - changes the points to a Cartesian coordinate system.
      */
    private int[] undoPoint(int[] XY) {
	
	// Based on the flags and order, fix the points back to the original quadrants.

	if ((flag & SWAP    ) == SWAP) XY = swap(XY[0], XY[1]);
	if ((flag & NEGATE_Y) == NEGATE_Y) XY[1] = neg(XY[1]);
	if ((flag & NEGATE_X) == NEGATE_X) XY[0] = neg(XY[0]);
	// Fix the points to the Java coordinate system.

	XY[0] = XY[0] + xOrigin;
	if ((XY[1] - height +  yOrigin) > 0) // This is only occurs when
	    XY[1] = -1;  // when have smaller numbers the 0 (in Java's coordinates)
	else 
	    XY[1] = Math.abs(XY[1] - height +  yOrigin);
	return XY;
    }
	
    /*
      * Draws a tiny little pixel (or a thick/patterned one depending on the state).
      * 
      * require:
      *  XY be in a Cartesian coordinate system.
      *
      */

    private void drawPoint (int[] XY) {

    

	if (maxPoints == nPoints) {

	    // Lots deleted ... 
	    if ((XY[0] > 0) || (XY[1] > 0)) {// No sense to draw negative points
	    // Just put a single point.
		OffG.drawRect(XY[0],XY[1],0,0);
	    
	    } else System.out.print(" o ");

	}
    }

    /*
      * Changes the sign of the array.
      * 
      * require:
      *   a.length() == 2
      *
      * note: if a[0] = -2. after execution, a[0] == 2.
      */
    private void neg (int[] a) {
	
	if ((a[0] > 0) || (a[1] > 0)) { a[0] = -a[0]; a[1]=-a[1];}
	    else { a[1] = Math.abs(a[1]); a[0] = Math.abs(a[0]); }
    }

    /*
      * Changes the sign of the integer.
      * 
      * require:
      *  a typeof Integer.
      *
      * return
      *  neg() typeof Integer
      */
    private int neg(int a) {

	if (a > 0) a = -a; else a = Math.abs(a);
	return a;
	
    }
    
    /* 
      * Swaps two points and put them in to an array.
      * 
      * require:
      *  a typeof Integer && b typeof Integer.
      *
      * return:
      *  swap().length == 2.
      *
      * note:
      *  swap(a,b).[0] == b; swap(a,b).[1] == a.
      */
    private int[] swap (int a, int b) {
    
	int c[] = new int[maxPoints];
	c[0] = b;
	c[1] = a;
	return c;
    }
    
    /*
      * Swaps the two arrays points
      *
      * require:
      *  a.length() == 2 && b.length() == 2.
      *
      */
    private void swap (int a[], int b[]) {
    
	int[] c = new int[maxPoints];
	c[0] = a[0];
	c[1] = a[1];
	a[0] = b[0];
	a[1] = b[1];
	b[0] = c[0];
	b[1] = c[1];
    }
    
    /*
      * Makes a copy of the array.
      *
      * require:
      *  a.length == 2;
      *
      * return:
      *  copy().lenght() == 2
      */
    private int[] copy (int[] a) {
    
	int[] b = new int[maxPoints];
	b[0] = a[0];
	b[1] = a[1];
	return b;
    }
    private int[] xPoints;
    private int[] yPoints;
    
    private Graphics OffG; // Where we will draw.
    private int xOrigin;  // Our origin's
    private int yOrigin;
    private int height;   
    private int width;
    private int maxPoints;
    private int nPoints;
    private int flag = 0;	// Did we swap, negate x, negate y the points?


}
