/*
 * Author:  Konrad Rzeszutek <conradus@mindless.com>
 * Date:    Oct 19th, 1999
 * File:    Editor3D/VectorObject.java
 *
 * Notes:
 * 
 *   This class models a 3D point representation.
 * The components of a vector are P^T = [x y z d], where d
 * is used b/c of homogonization.
 *
 */
 
package Editor3D;

public class VectorObject {

    // Couple of neccessary constants.
    
    public final static int X_POS = 0;
    public final static int Y_POS = 1;
    public final static int Z_POS = 2;
    public final static int W_POS = 3;
    // Yes, we could operate on 4D points.
    public final static int ROWS = 4;
    public final static int COLS = 4;
    
    /*
     * Initialize the VectorObject.
     *
     * require:
     *  x,y,z,w belong to R (R is real numbers).
     *
     */
    public VectorObject (double x, double y, double z, double w) {
    
	vec = new double[4];
	org = new double[4];
	temp = new double[4];
	wc = new double[4];
	
	org[X_POS] = wc[X_POS] = vec[X_POS] = x;
	org[Y_POS] = wc[Y_POS] = vec[Y_POS] = y;
	org[Z_POS] = wc[Z_POS] = vec[Z_POS] = z;
	org[W_POS] = wc[W_POS] = vec[W_POS] = w;
	
    }
    
    /*
     * Initalize the VectorObject.
     *
     * require:
     *   x,y,z belong to R (R is a set of real numbers)
     *
     */
    public VectorObject (double x, double y, double z) {
    
	this(x,y,z,1);
    }

    /*
     * Initalize the vector
     */
    public VectorObject () {
    
	this(0,0,0,1);
    }
    
    public boolean equal(VectorObject other) {
	
	return ( (vec[X_POS] == other.vec[X_POS]) && (vec[Y_POS] == other.vec[Y_POS]) && (vec[Z_POS] == other.vec[Z_POS]));
    }
    /*
     * Makes a copy of the vector.
     */
    public VectorObject copy() {
    
	return new VectorObject (vec[X_POS], vec[Y_POS], vec[Z_POS]);
    }
    /*
     * Sets the value of the WC (Word Co-ordinate) and X,Y,Z to the
     * original ones (the one submitted to the constructor).
     */    
    public void reset() {
    
	vec[X_POS] = wc[X_POS] = org[X_POS];
	vec[Y_POS] = wc[Y_POS] = org[Y_POS];
	vec[Z_POS] = wc[Z_POS] = org[Z_POS];
	vec[W_POS] = wc[W_POS] = org[W_POS];
    }
    
    /*
    * Multiply local points (X, Y, Z) against the matrix. 
    * Note, this does _NOT_ change the WC coordinates
    *
    * require:
    *   m.rows == m.columns == 4
    */
    public void multiply(Matrix m) {
    
	// We will multiply agains the m*P.
	// If you don't understand matrix multiplication, 
	// grab a Linear Algebra book and you will get the idea.
	
	for (int i = 0;i < ROWS;i++) {
	    temp[i] = 0.0; // First clear it.
	    for (int j = 0; j < COLS; j++)
		temp[i] += m.m[i][j] * vec[j]; 
	}
	// Copy it. We couldn't put the sum in vec[j], b/c we could have
	// screwed the values up for next computations
	
	for (int i = 0; i < ROWS; i++)
	    vec[i] = temp[i];
	
    }
    
    /*
     * Recalculate the WC using m matrix.
     * This doesn't change the X, Y, Z values. It only 'fixes'
     * the 'wc' array to the product of vec*m.
     *
     * require:
     *  m.rows == m.columns == 4
     *
     */     
    public void recalcWC(Matrix m) {
    

	for (int i = 0; i< ROWS;i++) {	    
	    wc[i] = 0.0;
	    for (int j =0 ; j < COLS; j++) 		
		wc[i] += m.m[i][j] * vec[j];
	    }
    }

    public void homogonize() {

	if (vec[W_POS] != 1.0) {
	//    System.out.println("Homogonize!");
	vec[X_POS] = vec[X_POS]/vec[W_POS];
	vec[Y_POS] = vec[Y_POS]/vec[W_POS];
	vec[Z_POS] = vec[Z_POS]/vec[W_POS];
	vec[W_POS] = vec[W_POS]/vec[W_POS];
	}
	if (wc[W_POS] != 1.0) {
//	System.out.print(".");

	wc[X_POS] = wc[X_POS]/wc[W_POS];
	wc[Y_POS] = wc[Y_POS]/wc[W_POS];
	wc[Z_POS] = wc[Z_POS]/wc[W_POS];
	wc[W_POS] = 1.0;
	}
    }
    /*
     * Cross product two VectorObject's.
     *
     * require:
     *   vec1 != null && vec2 != null
     */
    public static VectorObject crossProduct(VectorObject vec1, VectorObject vec2) {
    
	double newX = 0;
	double newY = 0;
	double newZ = 0;
	
	newX = vec1.vec[Y_POS]*vec2.vec[Z_POS] - vec1.vec[Z_POS]*vec2.vec[Y_POS];
	newY = vec1.vec[Z_POS]*vec2.vec[X_POS] - vec1.vec[X_POS]*vec2.vec[Z_POS];
	newZ = vec1.vec[X_POS]*vec2.vec[Y_POS] - vec1.vec[Y_POS]*vec2.vec[X_POS];
	
	return new VectorObject(newX, newY, newZ);
    }

 /*
     * Cross product two VectorObject's.
     *
     * require:
     *   vec1 != null && vec2 != null
     */
    public static VectorObject crossProduct2(VectorObject vec1, VectorObject vec2) {
    
	double newX = 0;
	double newY = 0;
	double newZ = 0;
	
	newX = vec1.wc[Y_POS]*vec2.wc[Z_POS] - vec1.wc[Z_POS]*vec2.wc[Y_POS];
	newY = vec1.wc[Z_POS]*vec2.wc[X_POS] - vec1.wc[X_POS]*vec2.wc[Z_POS];
	newZ = vec1.wc[X_POS]*vec2.wc[Y_POS] - vec1.wc[Y_POS]*vec2.wc[X_POS];
	
	return new VectorObject(newX, newY, newZ);
    }
    /*
     * Unit vector of a given vector
     *
     * require:
     *   vec != null
     */
    public static VectorObject unitVector(VectorObject vec) {
    
	double d = Math.sqrt( (vec.vec[X_POS]*vec.vec[X_POS]) +
			      (vec.vec[Y_POS]*vec.vec[Y_POS]) +
			      (vec.vec[Z_POS]*vec.vec[Z_POS]));
			      
	return new VectorObject(vec.vec[X_POS]/d, vec.vec[Y_POS]/d, vec.vec[Z_POS]/d);
    }

 /*
     * Unit vector of a given vector
     *
     * require:
     *   vec != null
     */
    public static VectorObject unitVector2(VectorObject vec) {
    
	double d = Math.sqrt( (vec.wc[X_POS]*vec.wc[X_POS]) +
			      (vec.wc[Y_POS]*vec.wc[Y_POS]) +
			      (vec.wc[Z_POS]*vec.wc[Z_POS]));
			      
	return new VectorObject(vec.wc[X_POS]/d, vec.wc[Y_POS]/d, vec.wc[Z_POS]/d);
    }

    /*
     * Perform a scalar multiplication on the given vector.
     *
     * require:
     *  num anything
     */
    public void scalarMultiply(double num) {

	this.vec[X_POS] = num*this.vec[X_POS];
	this.vec[Y_POS] = num*this.vec[Y_POS];
	this.vec[Z_POS] = num*this.vec[Z_POS];

	this.wc[X_POS] *= num;
	this.wc[Y_POS] *= num;
	this.wc[Z_POS] *= num;

    }

    /*
     * Perform vector addition 
     *
     * require:
     *  VectorObject instanceof null
     *
     */
    public void addVectors(VectorObject vec) {

	this.vec[X_POS] += vec.vec[X_POS];
	this.vec[Y_POS] += vec.vec[Y_POS];
	this.vec[Z_POS] += vec.vec[Z_POS];

	this.wc[X_POS] += vec.wc[X_POS];
	this.wc[Y_POS] += vec.wc[Y_POS];
	this.wc[Z_POS] += vec.wc[Z_POS];
    }

    /*
     * Subtract the vector points
     *
     * require:
     *  vec1 && vec2 == VectorObject.
     */
    public void subtractVectors(VectorObject vec1, VectorObject vec2) {

	this.vec[X_POS] = vec1.vec[X_POS] - vec2.vec[X_POS];
	this.vec[Y_POS] = vec1.vec[Y_POS] - vec2.vec[Y_POS];
	this.vec[Z_POS] = vec1.vec[Z_POS] - vec2.vec[Z_POS];


	this.wc[X_POS] = vec1.wc[X_POS] - vec2.wc[X_POS];
	this.wc[Y_POS] = vec1.wc[Y_POS] - vec2.wc[Y_POS];
	this.wc[Z_POS] = vec1.wc[Z_POS] - vec2.wc[Z_POS];

    }

    /*
     * Do a dot-product 
     *
     * require:
     *  vec instanceof VectorObject && vec instaceof VectorObject
     *
     * return:
     *  double.
     *
     */
    public static double dotProduct (VectorObject vec1, VectorObject vec2) {

	return  ((vec1.vec[X_POS]*vec2.vec[X_POS])+
		 (vec1.vec[Y_POS]*vec2.vec[Y_POS])+
		 (vec1.vec[Z_POS]*vec2.vec[Z_POS]));

    }
       /*
     * Do a dot-product 
     *
     * require:
     *  vec instanceof VectorObject && vec instaceof VectorObject
     *
     * return:
     *  double.
     *
     */
    public static double dotProduct2 (VectorObject vec1, VectorObject vec2) {

	return  ((vec1.wc[X_POS]*vec2.wc[X_POS])+
		 (vec1.wc[Y_POS]*vec2.wc[Y_POS])+
		 (vec1.wc[Z_POS]*vec2.wc[Z_POS]));

    }
     
     
    /*
     * Returns a nice string represenation.
     *
     * return:
     *   String.
     */
    public String toString () {
            
	return "P^T = [x: "+vec[X_POS]+" y: "+vec[Y_POS]+" z: "+vec[Z_POS]+" d: "+vec[W_POS]+"]\n"+
	       "      [x: "+wc[X_POS]+" y: "+wc[Y_POS]+" z: "+wc[Z_POS]+" d: "+wc[W_POS]+"]\n";
    }
    
    // World wide readable/changable variables.
    public double wc[];       // Water Closet :p, actually this stands for World Co-ordinate system
		       // Its used for Java X,Y drawing system.
    public double vec[];      // Our (UV) co-ordinate system.
    // Private values.
    private double org[];
    private double temp[];
    
} // End of VectorObject.
